# Generated by Django 2.2.5 on 2019-11-04 16:28

import asymmetric_jwt_auth.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('asymmetric_jwt_auth', '0003_auto_20151112_1547'),
    ]

    operations = [
        migrations.AddField(
            model_name='publickey',
            name='last_used_on',
            field=models.DateTimeField(blank=True, null=True, verbose_name='Last Used On'),
        ),
        migrations.AlterField(
            model_name='publickey',
            name='key',
            field=models.TextField(help_text="The user's RSA public key", validators=[asymmetric_jwt_auth.models.validate_public_key]),
        ),
    ]



from django.test import RequestFactory, TestCase
from django.contrib.auth.models import User
from asymmetric_jwt_auth import create_auth_header
from asymmetric_jwt_auth.models import PublicKey
from asymmetric_jwt_auth.middleware import JWTAuthMiddleware
from unittest import mock
import os.path

BASE = os.path.dirname(os.path.abspath(__file__))
KEY1_PRIVATE = os.path.join(BASE, 'dummy.privkey')
KEY1_PUBLIC = os.path.join(BASE, 'dummy.pub')


class MiddlewareTest(TestCase):

    def setUp(self):
        self.user = User.objects.create_user(username='foo')
        with open(KEY1_PUBLIC, 'r') as pubkey_file:
            key = pubkey_file.read()
        self.public_key = PublicKey.objects.create(
            user=self.user,
            key=key)


    def test_authenticate_request(self):
        header = create_auth_header('foo', key_file=KEY1_PRIVATE)
        rfactory = RequestFactory()
        request = rfactory.get('/', HTTP_AUTHORIZATION=header)

        next_middleware = mock.MagicMock()
        run_middleware = JWTAuthMiddleware(next_middleware)

        self.public_key.refresh_from_db()
        self.assertIsNone(self.public_key.last_used_on)
        self.assertEqual(next_middleware.call_count, 0)
        self.assertEqual(getattr(request, 'user', None), None)

        run_middleware(request)

        self.public_key.refresh_from_db()
        self.assertIsNotNone(self.public_key.last_used_on)
        self.assertEqual(next_middleware.call_count, 1)
        self.assertEqual(getattr(request, 'user', None), self.user)



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PrivateFormat,
    PublicFormat,
    NoEncryption
)
import os.path
import asymmetric_jwt_auth.token as token


default_app_config = 'asymmetric_jwt_auth.apps.JWTAuthConfig'


AUTH_METHOD = 'JWT'
ENCODING = 'utf-8'


def generate_key_pair(size=2048, public_exponent=65537, as_string=True):
    """
    Generate a public/private key pair.

    :param size: Optional. Describes how many bits long the key should be, larger keys provide more security,
        currently 1024 and below are considered breakable, and 2048 or 4096 are reasonable default
        key sizes for new keys. Defaults to 2048.
    :param public_exponent: Optional. Indicates what one mathematical property of the key generation will be.
        65537 is the default and should almost always be used.
    :param as_string: Optional. If True, return tuple of strings. If false, return tuple of RSA key objects.
        Defaults to True.
    :return: (PrivateKey<string>, PublicKey<string>)
    :return: (
        `RSAPrivateKey <https://cryptography.io/en/latest/hazmat/primitives/asymmetric/rsa/#cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey>`_,
        `RSAPublicKey <https://cryptography.io/en/latest/hazmat/primitives/asymmetric/rsa/#cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey>`_)
    """
    private = rsa.generate_private_key(
        public_exponent=public_exponent,
        key_size=size,
        backend=default_backend()
    )
    public = private.public_key()

    if not as_string:
        return private, public

    pem_private = private.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption()).decode(ENCODING)
    pem_public = public.public_bytes(Encoding.PEM, PublicFormat.SubjectPublicKeyInfo).decode(ENCODING)
    return pem_private, pem_public


def load_private_key(key_file, key_password=None):
    """
    Load a private key from disk.

    :param key_file: File path to key file.
    :param key_password: Optional. If the key file is encrypted, provide the password to decrypt it. Defaults to None.
    :return: PrivateKey<string>
    """
    key_file = os.path.expanduser(key_file)
    key_file = os.path.abspath(key_file)

    if not key_password:
        with open(key_file, 'r') as key:
            return key.read()

    with open(key_file, 'rb') as key:
        key_bytes = key.read()
    return decrypt_key(key_bytes, key_password).decode(ENCODING)


def decrypt_key(key, password):
    """
    Decrypt an encrypted private key.

    :param key: Encrypted private key as a string.
    :param password: Key pass-phrase.
    :return: Decrypted private key as a string.
    """
    private = serialization.load_pem_private_key(key, password=password, backend=default_backend())
    return private.private_bytes(Encoding.PEM, PrivateFormat.PKCS8, NoEncryption())


def create_auth_header(username, key=None, key_file="~/.ssh/id_rsa", key_password=None):
    """
    Create an HTTP Authorization header using a private key file.

    Either a key or a key_file must be provided.

    :param username: The username to authenticate as on the remote system.
    :param key: Optional. A private key as either a string or an instance of cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey.
    :param key_file: Optional. Path to a file containing the user's private key. Defaults to ~/.ssh/id_rsa. Should be in PEM format.
    :param key_password: Optional. Password to decrypt key_file. If set, should be a bytes object.
    :return: Authentication header value as a string.
    """
    if not key:
        key = load_private_key(key_file, key_password)
    claim = token.sign(username, key)
    return "%s %s" % (AUTH_METHOD, claim.decode(ENCODING))



from django.conf import settings
from django.db import models
from django.utils import timezone
from django.core.exceptions import ValidationError
from cryptography.hazmat.primitives.serialization import load_pem_public_key, load_ssh_public_key
from cryptography.hazmat.backends import default_backend


def validate_public_key(value):
    """
    Check that the given value is a valid RSA Public key in either PEM or OpenSSH format. If it is invalid,
    raises ``django.core.exceptions.ValidationError``.
    """
    is_valid = False
    exc = None

    for load in (load_pem_public_key, load_ssh_public_key):
        if not is_valid:
            try:
                load(value.encode('utf-8'), default_backend())
                is_valid = True
            except Exception as e:
                exc = e

    if not is_valid:
        raise ValidationError('Public key is invalid: %s' % exc)


class PublicKey(models.Model):
    """
    Store a public key and associate it to a particular user.

    Implements the same concept as the OpenSSH ``~/.ssh/authorized_keys`` file on a Unix system.
    """

    #: Foreign key to the Django User model. Related name: ``public_keys``.
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='public_keys', on_delete=models.CASCADE)

    #: Key text in either PEM or OpenSSH format.
    key = models.TextField(help_text="The user's RSA public key", validators=[validate_public_key])

    #: Comment describing the key. Use this to note what system is authenticating with the key, when it was last rotated, etc.
    comment = models.CharField(max_length=100, help_text="Comment describing this key", blank=True)

    #: Date and time that key was last used for authenticating a request.
    last_used_on = models.DateTimeField("Last Used On", null=True, blank=True)


    def update_last_used_datetime(self):
        self.last_used_on = timezone.now()
        self.save(update_fields=["last_used_on"])


    def save(self, *args, **kwargs):
        key_parts = self.key.split(' ')
        if len(key_parts) == 3 and not self.comment:
            self.comment = key_parts.pop()
        super(PublicKey, self).save(*args, **kwargs)



from django.contrib import admin
from . import models


@admin.register(models.PublicKey)
class PublicKeyAdmin(admin.ModelAdmin):
    list_display = ['user', 'comment', 'last_used_on']
    fields = ['user', 'comment', 'key', 'last_used_on']
    raw_id_fields = ['user']



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('asymmetric_jwt_auth', '0002_publickey_comment'),
    ]

    operations = [
        migrations.AlterField(
            model_name='publickey',
            name='comment',
            field=models.CharField(blank=True, max_length=100, help_text='Comment describing this key'),
        ),
    ]



import jwt
import time
import random
import logging

logger = logging.getLogger(__name__)


#: Default JWT signing algorithm
DEFAULT_ALGORITHM = 'RS512'

# Number of seconds of clock-drift to tolerate when verifying the authenticity of a JWT.
TIMESTAMP_TOLERANCE = 20  # Seconds


def sign(username, private_key, generate_nonce=None, iat=None, algorithm=DEFAULT_ALGORITHM):
    """
    Create a signed JWT using the given username and RSA private key.

    :param username: Username (string) to authenticate as on the remote system.
    :param private_key: Private key to use to sign the JWT claim.
    :param generate_nonce: Optional. Callable to use to generate a new nonce. Defaults to
        `random.random <https://docs.python.org/3/library/random.html#random.random>`_.
    :param iat: Optional. Timestamp to include in the JWT claim. Defaults to
        `time.time <https://docs.python.org/3/library/time.html#time.time>`_.
    :param algorithm: Optional. Algorithm to use to sign the JWT claim. Default to ``RS512``.
        See `pyjwt.readthedocs.io <https://pyjwt.readthedocs.io/en/latest/algorithms.html>`_ for other possible algorithms.
    :return: JWT claim as a string.
    """
    iat = iat if iat else time.time()
    if not generate_nonce:
        generate_nonce = lambda username, iat: random.random()  # NOQA

    token_data = {
        'username': username,
        'time': iat,
        'nonce': generate_nonce(username, iat),
    }

    token = jwt.encode(token_data, private_key, algorithm=algorithm)
    return token


def get_claimed_username(token):
    """
    Given a JWT, get the username that it is claiming to be `without verifying that the signature is valid`.

    :param token: JWT claim
    :return: Username
    """
    unverified_data = jwt.decode(token, options={
        'verify_signature': False
    })

    if 'username' not in unverified_data:
        return None
    return unverified_data['username']


def verify(token, public_key, validate_nonce=None, algorithms=[DEFAULT_ALGORITHM]):
    """
    Verify the validity of the given JWT using the given public key.

    :param token: JWM claim
    :param public_key: Public key to use when verifying the claim's signature.
    :param validate_nonce: Callable to use to validate the claim's nonce.
    :param algorithms: Allowable signing algorithms. Defaults to ['RS512'].
    :return: False if the token is determined to be invalid or a dictionary of the token data if it is valid.
    """
    try:
        token_data = jwt.decode(token, public_key, algorithms=algorithms)
    except jwt.InvalidTokenError:
        logger.debug('JWT failed verification')
        return False

    claimed_username = token_data.get('username')
    claimed_time = token_data.get('time', 0)
    claimed_nonce = token_data.get('nonce')

    # Ensure time is within acceptable bounds
    current_time = time.time()
    min_time, max_time = (current_time - TIMESTAMP_TOLERANCE, current_time + TIMESTAMP_TOLERANCE)
    if claimed_time < min_time or claimed_time > max_time:
        logger.debug('Claimed time is outside of allowable tolerances')
        return False

    # Ensure nonce is unique
    if validate_nonce:
        if not validate_nonce(claimed_username, claimed_time, claimed_nonce):
            logger.debug('Claimed nonce failed to validate')
            return False
    else:
        logger.warning('validate_nonce function was not supplied!')

    # If we've gotten this far, the token is valid
    return token_data



from django.test import TestCase
from asymmetric_jwt_auth import generate_key_pair
import asymmetric_jwt_auth.token as token
import time


class AuthTest(TestCase):

    def test_roundtrip(self):
        private, public = generate_key_pair()
        t = token.sign('guido', private)
        token_data = token.verify(t, public)
        self.assertTrue(token_data)
        self.assertEqual(token_data.get('username'), 'guido')


    def test_bad_keys(self):
        private1, public1 = generate_key_pair()
        private2, public2 = generate_key_pair()

        t = token.sign('guido', private1)
        token_data = token.verify(t, public1)
        self.assertTrue(token_data)

        t = token.sign('guido', private2)
        token_data = token.verify(t, public2)
        self.assertTrue(token_data)

        t = token.sign('guido', private1)
        token_data = token.verify(t, public2)
        self.assertFalse(token_data)

        t = token.sign('guido', private2)
        token_data = token.verify(t, public1)
        self.assertFalse(token_data)


    def test_bad_iat(self):
        private, public = generate_key_pair()

        t = token.sign('guido', private, iat=time.time())
        token_data = token.verify(t, public)
        self.assertTrue(token_data)

        # IAT tolerance exists to account for clock drift between disparate systems.
        tolerance = token.TIMESTAMP_TOLERANCE + 1

        t = token.sign('guido', private, iat=time.time() - tolerance)
        token_data = token.verify(t, public)
        self.assertFalse(token_data)

        t = token.sign('guido', private, iat=time.time() + tolerance)
        token_data = token.verify(t, public)
        self.assertFalse(token_data)


    def test_bad_nonce(self):
        private, public = generate_key_pair()

        t = token.sign('guido', private, generate_nonce=lambda username, iat: 1)
        token_data = token.verify(t, public, validate_nonce=lambda username, iat, nonce: nonce == 1)
        self.assertTrue(token_data)

        t = token.sign('guido', private, generate_nonce=lambda username, iat: 1)
        token_data = token.verify(t, public, validate_nonce=lambda username, iat, nonce: nonce == 2)
        self.assertFalse(token_data)

        t = token.sign('guido', private, generate_nonce=lambda username, iat: 2)
        token_data = token.verify(t, public, validate_nonce=lambda username, iat, nonce: nonce == 1)
        self.assertFalse(token_data)


    def test_get_claimed_username(self):
        private, public = generate_key_pair()
        t = token.sign('guido', private)
        self.assertEqual(token.get_claimed_username(t), 'guido')



from django.test import TestCase
import asymmetric_jwt_auth as auth


class UtilsTest(TestCase):
    def test_generate_new_key(self):
        private, public = auth.generate_key_pair()

        private = private.strip().split('\n')
        self.assertEqual(private[0], '-----BEGIN PRIVATE KEY-----')
        self.assertEqual(private[27], '-----END PRIVATE KEY-----')

        public = public.strip().split('\n')
        self.assertEqual(public[0], '-----BEGIN PUBLIC KEY-----')
        self.assertEqual(public[8], '-----END PUBLIC KEY-----')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PublicKey',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, verbose_name='ID', primary_key=True)),
                ('key', models.TextField(help_text="The user's RSA public key")),
                ('user', models.ForeignKey(to=settings.AUTH_USER_MODEL, related_name='public_keys', on_delete=models.CASCADE)),
            ],
        ),
    ]



#!/usr/bin/env python
from setuptools import setup, find_packages, Distribution
import codecs
import os.path

# Make sure versiontag exists before going any further
Distribution().fetch_build_eggs('versiontag>=1.2.0')

from versiontag import get_version, cache_git_tag  # NOQA


packages = find_packages('src')

install_requires = [
    'PyJWT>=1.4.2,<2.0.0',
    'cryptography>=1.7.1',
    'Django>=2.2',
]

extras_require = {
    'development': [
        'flake8>=3.2.1',
        'sphinx>=1.5.2',
        'versiontag>=1.2.0',
    ],
}


def fpath(name):
    return os.path.join(os.path.dirname(__file__), name)


def read(fname):
    return codecs.open(fpath(fname), encoding='utf-8').read()


cache_git_tag()

setup(
    name='asymmetric_jwt_auth',
    description='Asymmetric key based authentication for HTTP APIs',
    version=get_version(pypi=True),
    long_description=read(fpath('README.rst')),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: Unix',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    author='Craig Weber',
    author_email='crgwbr@gmail.com',
    url='https://github.com/crgwbr/asymmetric-jwt-auth',
    license='ISC',
    package_dir={'': 'src'},
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require
)



from django.contrib.auth import get_user_model
from django.core.cache import cache
from asymmetric_jwt_auth import AUTH_METHOD
import asymmetric_jwt_auth.token as token
import logging

logger = logging.getLogger(__name__)


class JWTAuthMiddleware(object):
    """Django middleware class for authenticating users using JWT Authentication headers"""

    def __init__(self, get_response):
        self.get_response = get_response


    def __call__(self, request):
        # Attempt to authorize the request
        self.authorize_request(request)
        # Continue with the request
        return self.get_response(request)


    def authorize_request(self, request):
        """
        Process a Django request and authenticate users.

        If a JWT authentication header is detected and it is determined to be valid, the user is set as
        ``request.user`` and CSRF protection is disabled (``request._dont_enforce_csrf_checks = True``) on
        the request.

        :param request: Django Request instance
        """
        if 'HTTP_AUTHORIZATION' not in request.META:
            return request

        try:
            method, claim = request.META['HTTP_AUTHORIZATION'].split(' ', 1)
        except ValueError:
            return request

        if method.upper() != AUTH_METHOD:
            return request

        username = token.get_claimed_username(claim)
        if not username:
            return request

        User = get_user_model()
        try:
            user = User.objects.get(username=username)
        except User.DoesNotExist:
            return request

        claim_data = None
        for public in user.public_keys.all():
            claim_data = token.verify(claim, public.key, validate_nonce=self.validate_nonce)
            if claim_data:
                public.update_last_used_datetime()
                break
        if not claim_data:
            return request

        logger.debug('Successfully authenticated %s using JWT', user.username)
        request._dont_enforce_csrf_checks = True
        request.user = user
        return request


    def create_nonce_key(self, username, iat):
        """
        Create and return the cache key for storing nonces

        :param username: Username as a string.
        :param iat: Unix timestamp float or integer of when the nonce was used.
        :return: Cache key string.
        """
        return '%s-nonces-%s-%s' % (
            self.__class__.__name__,
            username,
            iat,
        )


    def log_used_nonce(self, username, iat, nonce):
        """
        Log a nonce as being used, and therefore henceforth invalid.

        :param username: Username as a string.
        :param iat: Unix timestamp float or integer of when the nonce was used.
        :param nonce: Nonce value.
        """
        # TODO: Figure out some way to do this in a thread-safe manner. It'd be better to use
        # a Redis Set or something, but we don't necessarily want to be tightly coupled to
        # Redis either since not everyone uses it.
        key = self.create_nonce_key(username, iat)
        used = cache.get(key, [])
        used.append(nonce)
        cache.set(key, set(used), token.TIMESTAMP_TOLERANCE * 2)


    def validate_nonce(self, username, iat, nonce):
        """
        Confirm that the given nonce hasn't already been used.

        :param username: Username as a string.
        :param iat: Unix timestamp float or integer of when the nonce was used.
        :param nonce: Nonce value.
        :return: True if nonce is valid, False if it is invalid.
        """
        key = self.create_nonce_key(username, iat)
        used = cache.get(key, [])
        return nonce not in used
